#[macro_use]
extern crate lazy_static;
extern crate chrono;
extern crate iron;
extern crate rand;
extern crate sequoia_openpgp as openpgp;
extern crate snap;

mod pgp;

use crate::pgp::KnownKeys;
use byteorder::{BigEndian, ReadBytesExt, WriteBytesExt};
use chrono::*;
use core::cell::RefCell;
use iron::method::Method;
use iron::modifiers::Redirect;
use iron::prelude::*;
use iron::url::Url;
use rand::Rng;
use std::borrow::BorrowMut;
use std::fs;
use std::fs::File;
use std::io;
use std::io::prelude::*;
use std::iter::Iterator;
use std::path::{Path, PathBuf};
use std::sync::Mutex;

struct Snippet<'a> {
    id: String,
    storage: &'a SnippetStorage<'a>,
}

#[allow(dead_code)]
struct SnippetMeta {
    created: DateTime<Utc>,
    compression: Option<String>,
}

struct SnippetStorage<'a> {
    directory: &'a Path,
}

impl<'a> Snippet<'a> {
    pub fn random(storage: &'a SnippetStorage) -> Snippet<'a> {
        Snippet::new(
            &rand::thread_rng()
                .gen_ascii_chars()
                .take(6)
                .collect::<String>(),
            storage,
        )
    }

    pub fn new(id: &str, storage: &'a SnippetStorage) -> Snippet<'a> {
        Snippet {
            id: id.to_string(),
            storage,
        }
    }

    pub fn path(&self) -> PathBuf {
        self.storage.directory.join(&self.id)
    }

    pub fn metadata(&self) -> Result<SnippetMeta, io::Error> {
        let mut file = File::open(self.path())?;
        self.metadata_via_handle(&mut file)
    }

    fn metadata_via_handle(&self, hdl: &mut impl Read) -> Result<SnippetMeta, io::Error> {
        let timestamp = hdl.read_i64::<BigEndian>()?;
        let comp_len = hdl.read_u16::<BigEndian>()? as usize;
        let mut comp = Vec::with_capacity(comp_len);
        comp.resize(comp_len, 0u8);
        hdl.read_exact(&mut comp)?;
        let comp = String::from_utf8(comp).unwrap();
        Ok(SnippetMeta {
            created: Utc.timestamp_millis(timestamp),
            compression: Some(comp).filter(|_| comp_len > 0),
        })
    }

    fn contents(&self) -> Result<String, io::Error> {
        let mut file = File::open(self.path())?;
        let meta = self.metadata_via_handle(&mut file)?;
        fn read_string(r: &mut impl Read) -> Result<String, io::Error> {
            let mut text = String::new();
            r.read_to_string(&mut text)?;
            Ok(text)
        }
        dbg!((&meta.compression, &meta.created));
        match meta.compression {
            Some(ref comp) if comp == "snap" => {
                let mut r = snap::Reader::new(&mut file);
                read_string(&mut r)
            }
            _ => read_string(&mut file),
        }
    }

    fn write(self, content: &str) -> Result<Snippet<'a>, io::Error> {
        let mut file = File::create(self.path())?;
        file.write_i64::<BigEndian>(Utc::now().timestamp())?;
        let comp = if content.len() > 2048 {
            Some("snap")
        } else {
            None
        };
        file.write_u16::<BigEndian>(comp.map(|s| s.len() as u16).unwrap_or(0u16))?;
        file.write(comp.map(|s| s.as_bytes()).unwrap_or(&[0u8; 0]))?;
        match comp {
            Some(ref comp) if comp == &"snap" => {
                let mut w = snap::Writer::new(&mut file);
                w.write_all(content.as_bytes())?
            }
            _ => file.write_all(content.as_bytes())?,
        };
        Ok(Snippet::new(&self.id, self.storage))
    }
}

impl<'a> SnippetStorage<'a> {
    pub fn new(directory: &'a Path) -> SnippetStorage<'a> {
        SnippetStorage {
            directory: directory,
        }
    }

    fn has(&self, id: &str) -> bool {
        self.directory.join(id).exists()
    }

    fn open(&self, id: &str) -> Option<Snippet> {
        if self.has(id) {
            Some(Snippet::new(id, self))
        } else {
            None
        }
    }
}

#[cfg(not(debug_assertions))]
const STORAGE_DIR: &str = "/snips";

#[cfg(debug_assertions)]
const STORAGE_DIR: &str = "/tmp";

lazy_static! {
    static ref KNOWN_KEYS: Mutex<KnownKeys> = Mutex::new(
        KnownKeys::load_dir([STORAGE_DIR, "keys"].join("/")).expect("Failed to load pubkeys")
    );
}

const VERSION: &str = env!("CARGO_PKG_VERSION");

fn handle(req: &mut Request) -> IronResult<Response> {
    println!("{}", req.url);
    let storage = SnippetStorage::new(&Path::new(STORAGE_DIR));
    let segments: Vec<&str> = req.url.path();
    match (&req.method, segments.first()) {
        (Method::Get, Some(&"version")) => Ok(Response::with((iron::status::Ok, VERSION))),
        (Method::Post, Some(path)) => {
            if path == &"new" {
                let snip = {
                    let pgp_text: String = {
                        let bytes = ((&mut req.body).bytes().take(1024 * 512).collect::<Result<
                            Vec<u8>,
                            io::Error,
                        >>(
                        ))
                        .map_err(|err| IronError::new(err, ""))?;
                        String::from_utf8(bytes)
                            .map_err(|err| IronError::new(err, "Invalid utf8"))?
                    };
                    let b_text = KNOWN_KEYS
                        .lock().unwrap()//.map_err(|_| IronError::new(std::error::Error::from("Mutex Err"), "PGP Context unavailable"))?
                        .verify(pgp_text.as_bytes())
                        .map_err(|err| IronError::new(err, "Untrusted signature"))?;
                    let text = String::from_utf8(b_text).unwrap();
                    Snippet::random(&storage).write(&*text).map_err(|err| {
                        let msg = format!("Failed to save snippet: {:?}", &err);
                        IronError::new(err, msg)
                    })
                };
                snip.map(|snip| {
                    let mut snip_url: Url = req.url.clone().into();
                    snip_url.set_path(&*("/".to_string() + &*snip.id));
                    Response::with((
                        iron::status::TemporaryRedirect,
                        Redirect(iron::Url::from_generic_url(snip_url).unwrap()),
                    ))
                    /*Response::with((
                        iron::status::Ok,
                        format!(
                            "<meta http-equiv=\"refresh\" content=\"0; url={}/\" />",
                            snip_url
                        ),
                    ))*/
                })
            } else {
                Ok(Response::with((
                    iron::status::BadRequest,
                    "Post to /new or die",
                )))
            }
        }
        (Method::Get, Some(id)) => {
            let att = storage.open(&id).map(|snip| snip.contents()).map(|res| {
                Response::with(
                    match res.map(|text| (iron::status::Ok, text)).map_err(|err| {
                        let msg = format!("Failed to load snippet: {:?}", &err);
                        msg
                    }) {
                        Ok(res) => res,
                        Err(e) => (iron::status::InternalServerError, e),
                    },
                )
            });
            Ok(att.unwrap_or(Response::with((iron::status::NotFound, "Not here sry"))))
        }
        (Method::Get, _) => Ok(Response::with((iron::status::NotFound, "Wrong path pal"))),
        _ => Ok(Response::with((iron::status::BadRequest, "Give or take"))),
    }
}

fn main() {
    let chain = Chain::new(handle);
    println!("Starting brownpaper: {}", STORAGE_DIR);
    Iron::new(chain).http("0.0.0.0:3000").unwrap();
}
